from typing import Optional
from pydantic import BaseModel


class Sandbox(BaseModel):
    """
    Represents a sandbox for running AI-generated code and commands,
    as defined in the API spec.

    Attributes:
        id: The unique identifier of the sandbox.
        image: The container image the sandbox will run with.
    """

    id: Optional[str] = None
    image: str = ""


class RunIPythonCellRequest(BaseModel):
    """
    Represents the cell to run in a stateful IPython (Jupyter) kernel,
    as defined in the API spec.

    Attributes:
        code: The code to run in the IPython kernel.
        split_output: Whether to split output into stdout and stderr.
    """

    code: str
    split_output: Optional[bool] = None


class RunIPythonCellResponse(BaseModel):
    """
    Represents the response from a stateful IPython (Jupyter) kernel,
    as defined in the API spec.

    Attributes:
        output: The combined stdout and stderr from the IPython kernel.
        stdout: The stdout stream from the IPython kernel.
        stderr: The stderr stream from the IPython kernel.
    """

    output: Optional[str] = None
    stdout: Optional[str] = None
    stderr: Optional[str] = None


class RunShellCommandRequest(BaseModel):
    """
    Represents a shell command to run in the sandbox,
    as defined in the API spec.

    Attributes:
        command: The command to execute. Will be passed to 'bash -c'.
        split_output: Whether to split output into stdout and stderr.
                     If True, the output field in the response will be empty and
                     stdout/stderr fields will be populated.
    """

    command: str
    split_output: Optional[bool] = None


class RunShellCommandResponse(BaseModel):
    """
    Represents the response from a shell command execution,
    as defined in the API spec.

    Attributes:
        output: The combined stdout and stderr from the shell command.
        stdout: The stdout stream from the shell command.
        stderr: The stderr stream from the shell command.
    """

    output: Optional[str] = None
    stdout: Optional[str] = None
    stderr: Optional[str] = None


class Error(BaseModel):
    """
    Represents an API error response, as defined in the API spec.

    Attributes:
        message: The error message describing what went wrong.
    """

    message: str



from sandboxai.client.v1 import HttpClient
from sandboxai.api import v1 as v1Api
from sandboxai.utils import launch_sandboxd


DEFAULT_BACKEND = "http://localhost:5266/v1"
DEFAULT_IMAGE = "substratusai/sandboxai-box:v0.1.0"


class Sandbox:
    def __init__(
        self,
        backend: str = DEFAULT_BACKEND,
        image: str = DEFAULT_IMAGE,
        lazy_start: bool = False,
    ):
        """
        Initialize a Sandbox instance.
        """
        self.id = ""
        self.image = image
        self.backend = backend
        self.client = HttpClient(self.backend)
        if not self.client.check_health():
            if backend == DEFAULT_BACKEND:
                launch_sandboxd()
                self.client.wait_until_healthy(timeout=10)
            else:
                raise RuntimeError(
                    f"Sandbox service {backend}/v1/healthz is not responding with 200 OK."
                )
        if lazy_start == False:
            self.start()

    def __enter__(self):
        """
        Enter the context manager. Ensures the sandbox is started.
        """
        if not self.id:
            self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Exit the context manager. Stops the sandbox.
        """
        self.stop()
        return False  # Don't suppress any exceptions

    def start(self) -> None:
        response = self.client.create_sandbox(v1Api.Sandbox(image=self.image))
        self.id = response.id
        self.image = response.image

    def run_python_code(self, code: str) -> str:
        """
        Runs Python code in the sandbox.
        """
        if not self.id:
            self.start()

        response = self.client.run_ipython_cell(self.id, v1Api.RunIPythonCellRequest(code=code, split_output=False))  # type: ignore
        return response.output or ""

    def run_shell_command(self, command: str) -> str:
        """
        Runs a shell command in the sandbox (uses 'bash -c <command>').
        """
        if not self.id:
            self.start()

        response = self.client.run_shell_command(self.id, v1Api.RunShellCommandRequest(command=command, split_output=False))  # type: ignore
        return response.output or ""

    def stop(self) -> None:
        if self.id:
            self.client.delete_sandbox(self.id)
            self.id = ""
            self.image = ""



import sys
import shutil
import subprocess
import logging


def launch_sandboxd() -> int:
    """
    Launches the sandboxd binary and returns the process ID.
    """
    if not shutil.which("docker"):
        raise RuntimeError("docker not found on the system.")

    # Check if sandboxd binary is on the system.
    # TODO automatically download sandboxd binary if not present.
    if not shutil.which("sandboxd"):
        raise RuntimeError("sandboxd binary not found on the system.")

    # Launch the sandboxd binary in the background
    process = subprocess.Popen(
        ["sandboxd"],
        # TODO(nstogner): Log to disk.
        stdout=sys.stdout,
        stderr=sys.stderr,
        text=True,
    )

    # Check if the process started successfully
    if process.poll() is None:
        logging.info("sandboxd started successfully with PID: %d", process.pid)
        return process.pid
    else:
        raise RuntimeError("Failed to launch sandboxd")



import os
import pytest

from sandboxai.api.v1 import Sandbox, RunIPythonCellRequest, RunShellCommandRequest
from sandboxai.client.v1 import HttpClient


@pytest.fixture
def client():
    # Adjust the base_url as needed for your environment
    base_url = os.environ.get("SANDBOXAI_BASE_URL", "")
    return HttpClient(base_url=base_url)


def test_http_client_v1(client):
    """
    Run IPython tool end-to-end test for the Client.

    This test verifies that the client can create a sandbox, retrieve it,
    and run multiple IPython cells and shell command with the expected outputs.
    """
    box_image = os.environ.get("BOX_IMAGE", "")
    # Create a new sandbox (adjust the image as needed for your tests)
    sandbox = client.create_sandbox(Sandbox(image=box_image))

    # Ensure sandbox retrieval works
    retrieved = client.get_sandbox(sandbox.id)
    assert retrieved.id == sandbox.id

    ipy_test_cases = [
        {
            "name": "simple print",
            "code": "print(123)",
            "split": False,
            "expectedOutput": "123\n",
        },
        {
            "name": "simple split print",
            "code": "print(123)",
            "split": True,
            "expectedStdout": "123\n",
        },
        {
            "name": "import sys",
            "code": "import sys",
            # No expected output
        },
        {
            "name": "stderr print",
            "code": "import sys\nprint(123, file=sys.stderr)",
            "split": False,
            "expectedOutput": "123\n",
        },
        {
            "name": "stderr split print",
            "code": "import sys\nprint(123, file=sys.stderr)",
            "split": True,
            "expectedStderr": "123\n",
        },
        {
            "name": "unknown var",
            "code": "foo",
            "expectedOutputContains": "name 'foo' is not defined",
        },
        {
            "name": "set var",
            "code": "foo = 123",
        },
        {
            "name": "evaluate var",
            "code": "foo",
            "expectedOutput": "Out[1]: 123\n",
        },
    ]

    shell_test_cases = [
        {
            "name": "basic echo",
            "command": "echo 'hello'",
            "expectedOutput": "hello\n",
        },
        {
            "name": "echo to stderr",
            "command": ">&2 echo 'error'",
            "expectedOutput": "error\n",
        },
        {
            "name": "echo to stderr with split output",
            "command": ">&2 echo 'error'",
            "split": True,
            "expectedStderr": "error\n",
        },
    ]

    try:
        for tc in ipy_test_cases:
            req = RunIPythonCellRequest(
                code=tc["code"], split_output=tc.get("split", False)
            )
            resp = client.run_ipython_cell(sandbox.id, req)

            # If we expect the output to contain a substring
            if "expectedOutputContains" in tc:
                assert tc["expectedOutputContains"] in (resp.output or "")
                # This check ensures there's no conflicting expectations
                assert (
                    "expectedOutput" not in tc
                ), "Invalid assertion combo: both 'output' and 'outputContains' set."
            else:
                # Must match exact output if 'expectedOutput' is defined
                expected_output = tc.get("expectedOutput", "")
                assert expected_output == (resp.output or "")

                # Check stdout if provided
                if "expectedStdout" in tc:
                    assert tc["expectedStdout"] == (resp.stdout or "")

                # Check stderr if provided
                if "expectedStderr" in tc:
                    assert tc["expectedStderr"] == (resp.stderr or "")
        for tc in shell_test_cases:
            req = RunShellCommandRequest(
                command=tc["command"], split_output=tc.get("split", False)
            )
            resp = client.run_shell_command(sandbox.id, req)

            if "expectedOutput" in tc:
                assert tc["expectedOutput"] == (resp.output or "")
            if "expectedStdout" in tc:
                assert tc["expectedStdout"] == (resp.stdout or "")
            if "expectedStderr" in tc:
                assert tc["expectedStderr"] == (resp.stderr or "")

    finally:
        # Cleanup: delete the sandbox
        client.delete_sandbox(sandbox.id)
